/*
 *      scriptdialog.cpp
 *
 *      Copyright 2014 David Vachulka <arch_dvx@users.sourceforge.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

//#include <QPalette>
#include "scriptdialog.h"
#include "config.h"
#include "i18n.h"
#include "dxirc.h"

ScriptHighlighter::ScriptHighlighter(QTextDocument *parent)
    : QSyntaxHighlighter(parent)
{
    HighlightingRule rule;
    m_keywordFormat.setForeground(Qt::darkBlue);
    m_keywordFormat.setFontWeight(QFont::Bold);
    QStringList keywordPatterns;
    keywordPatterns << "\\bbreak\\b" << "\\bdo\\b" << "\\bend\\b"
                    << "\\belse\\b" << "\\belseif\\b" << "\\bfunction\\b"
                    << "\\bif\\b" << "\\blocal\\b" << "\\bnil\\b"
                    << "\\bnot\\b" << "\\bor\\b" << "\\brepeat\\b"
                    << "\\breturn\\b" << "\\bthen\\b" << "\\buntil\\b"
                    << "\\bwhile\\b";
    foreach(QString pattern, keywordPatterns)
    {
        rule.pattern = QRegExp(pattern);
        rule.format = m_keywordFormat;
        m_highlightingRules.append(rule);
    }
    m_singleLineCommentFormat.setForeground(Qt::red);
    rule.pattern = QRegExp("--[^\n]*");
    rule.format = m_singleLineCommentFormat;
    m_highlightingRules.append(rule);
    m_quotationFormat.setForeground(Qt::darkGreen);
    rule.pattern = QRegExp("\".*\"");
    rule.format = m_quotationFormat;
    m_highlightingRules.append(rule);
    m_functionFormat.setFontItalic(true);
    m_functionFormat.setForeground(Qt::blue);
    rule.pattern = QRegExp("\\b[A-Za-z0-9_]+(?=\\()");
    rule.format = m_functionFormat;
    m_highlightingRules.append(rule);
}

void ScriptHighlighter::highlightBlock(const QString &text)
{
    foreach(HighlightingRule rule, m_highlightingRules)
    {
        QRegExp expression(rule.pattern);
        int index = text.indexOf(expression);
        while(index >= 0)
        {
            int length = expression.matchedLength();
            setFormat(index, length, rule.format);
            index = text.indexOf(expression, index + length);
        }
    }
}

ScriptViewDialog::ScriptViewDialog(const QString& filename, QWidget* parent)
    : QDialog(parent)
{
    QVBoxLayout *mainframe = new QVBoxLayout;
    QFont font;
    font.setFamily("Courier");
    font.setFixedPitch(true);
    font.setPointSize(10);
    m_editor = new QTextEdit;
    m_editor->setReadOnly(true);
    m_editor->setFont(font);
    m_highlighter = new ScriptHighlighter(m_editor->document());
    QFile file(filename);
    if(file.open(QFile::ReadOnly | QFile::Text))
        m_editor->setPlainText(QString::fromUtf8(file.readAll()));
    QDialogButtonBox *buttons = new QDialogButtonBox(QDialogButtonBox::Ok);
    connect(buttons, SIGNAL(accepted()), this, SLOT(accept()));
    mainframe->addWidget(m_editor);
    mainframe->addWidget(buttons);
    setLayout(mainframe);
    setWindowTitle(filename.section(QDir::separator(), -1));
    resize(600,400);
}

ScriptDialog::ScriptDialog(QWidget *parent)
    : QDialog(parent)
{
    QVBoxLayout *contents = new QVBoxLayout;

    QHBoxLayout *scriptframe = new QHBoxLayout;
    m_names = new QListWidget;
    connect(m_names, SIGNAL(currentRowChanged(int)), this, SLOT(slotList()));
    scriptframe->addWidget(m_names);

    QGroupBox *group = new QGroupBox(_("Details"));
    QGridLayout *matrix = new QGridLayout;
    matrix->addWidget(new QLabel(_("Name:")), 0, 0);
    m_name = new QLabel;
    m_name->setFrameStyle(QFrame::Sunken|QFrame::Panel);
    matrix->addWidget(m_name, 0, 1);
    matrix->addWidget(new QLabel(_("Version:")), 1, 0);
    m_version = new QLabel;
    m_version->setFrameStyle(QFrame::Sunken|QFrame::Panel);
    matrix->addWidget(m_version, 1, 1);
    matrix->addWidget(new QLabel(_("Description:")), 2, 0);
    m_description = new QLabel;
    m_description->setFrameStyle(QFrame::Sunken|QFrame::Panel);
    matrix->addWidget(m_description, 2, 1);
    matrix->addWidget(new QLabel(_("Path:")), 3, 0);
    m_path = new QLabel;
    m_path->setFrameStyle(QFrame::Sunken|QFrame::Panel);
    matrix->addWidget(m_path, 3, 1);
    group->setLayout(matrix);
    scriptframe->addWidget(group);

    QDialogButtonBox *buttons = new QDialogButtonBox(QDialogButtonBox::Ok);
    QPushButton *buttonLoad = new QPushButton(_("L&oad"));
    connect(buttonLoad, SIGNAL(clicked()), this, SLOT(slotLoad()));
    buttons->addButton(buttonLoad, QDialogButtonBox::ActionRole);
    m_buttonView = new QPushButton(_("&View current"));
    connect(m_buttonView, SIGNAL(clicked()), this, SLOT(slotView()));
    buttons->addButton(m_buttonView, QDialogButtonBox::ActionRole);
    QPushButton *buttonViewNew = new QPushButton(_("View &script"));
    connect(buttonViewNew, SIGNAL(clicked()), this, SLOT(slotViewNew()));
    buttons->addButton(buttonViewNew, QDialogButtonBox::ActionRole);
    m_buttonUnload = new QPushButton(_("&Unload"));
    connect(m_buttonUnload, SIGNAL(clicked()), this, SLOT(slotUnload()));
    buttons->addButton(m_buttonUnload, QDialogButtonBox::ActionRole);
    connect(buttons, SIGNAL(accepted()), this, SLOT(accept()));
    contents->addLayout(scriptframe);
    contents->addWidget(buttons);
    setLayout(contents);
    setWindowTitle(_("Scripts list"));

    updateList();
    updateDetails();
}

void ScriptDialog::slotLoad()
{
    QFileDialog::Options options = QFileDialog::DontResolveSymlinks;
    QString path = QFileDialog::getOpenFileName(this,
                                _("Load lua script"),
                                "",
                                _("Lua scripts *.lua"),
                                NULL,
                                options);
    if(!path.isEmpty())
    {
        dxirc::instance()->loadLuaScript(path);
    }
    updateList();
    updateDetails();
}

void ScriptDialog::slotView()
{
    int index = m_names->currentRow();
    if(index == -1) return;
    viewFile(dxirc::instance()->getScripts()[index].path);
}

void ScriptDialog::slotViewNew()
{
    QFileDialog::Options options = QFileDialog::DontResolveSymlinks;
    QString path = QFileDialog::getOpenFileName(this,
                                _("Load lua script"),
                                "",
                                _("Lua scripts *.lua"),
                                NULL,
                                options);
    if(!path.isEmpty())
    {
        viewFile(path);
    }
}

void ScriptDialog::slotUnload()
{
    int index = m_names->currentRow();
    if(index == -1) return;
    dxirc::instance()->unloadLuaScript(dxirc::instance()->getScripts()[index].name);
    updateList();
    updateDetails();
}

void ScriptDialog::slotList()
{
    updateDetails();
}

void ScriptDialog::updateList()
{
    m_names->clear();
    for(int i=0; i<dxirc::instance()->getScripts().size(); i++)
    {
        m_names->addItem(dxirc::instance()->getScripts()[i].name);
    }
    if(dxirc::instance()->getScripts().size())
    {
        m_buttonView->setEnabled(true);
        m_buttonUnload->setEnabled(true);
    }
    else
    {
        m_buttonView->setEnabled(false);
        m_buttonUnload->setEnabled(false);
    }
    if(m_names->count()) m_names->setCurrentRow(m_names->count()-1);
}

void ScriptDialog::updateDetails()
{
    int index = m_names->currentRow();
    if(index == -1)
    {
        m_name->setText("");
        m_version->setText("");
        m_description->setText("");
        m_path->setText("");
        return;
    }
    m_name->setText(dxirc::instance()->getScripts()[index].name);
    m_version->setText(dxirc::instance()->getScripts()[index].version);
    m_description->setText(dxirc::instance()->getScripts()[index].description);
    m_path->setText(dxirc::instance()->getScripts()[index].path);
}

void ScriptDialog::viewFile(const QString &file)
{
    ScriptViewDialog *dialog = new ScriptViewDialog(file, this);
    dialog->setModal(true);
    dialog->show();
}

