/*
 *      dxutils.cpp
 *
 *      Copyright 2014 David Vachulka <arch_dvx@users.sourceforge.net>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#ifdef WIN32
 #define WIN32_LEAN_AND_MEAN
 #include <windows.h>
 #include <mmsystem.h>
 #include <ctime>
 #include <shellapi.h>
#endif
#include <QApplication>
#include <QColor>
#include <QBuffer>
#include <QFile>
#include <QDebug>
#include <QTime>
#include <QDir>
#include <QRegExp>
#include <QProcessEnvironment>
#include <QSettings>
#include <QSound>
#include <string>
#include "preferences.h"
#include "dxutils.h"
#include "notify/notifywindow.h"
#include "i18n.h"
#ifdef HAVE_ENCHANT
#include <enchant++.h>
#endif
#ifdef Q_OS_DARWIN
 #include "widgets/macspellcheck.h"
#endif

#ifdef HAVE_ENCHANT
typedef std::pair<Checkers*,QStringList*> CheckerPair;
void enumerateDicts (const char * const lang_tag,
         const char * const /*provider_name*/,
         const char * const /*provider_desc*/,
         const char * const /*provider_file*/,
         void * user_data)
{
    CheckerPair *chpair = (CheckerPair*)user_data;
    const Checkers &checkers = *chpair->first;
    QStringList &spellLang = *(chpair->second);
    if(checkers.find(lang_tag) == checkers.end())
        spellLang.append(lang_tag);
}
#endif

#ifdef Q_OS_DARWIN
static MacSpellCheck *_instance;
#endif

bool caseInsensitiveLessThan(const QString &s1, const QString &s2)
{
    return s1.toLower() < s2.toLower();
}

dxutils::dxutils()
{
    m_notifiesHeight = 0;
    m_notifyPosition = -1;
    m_firstNotify = false;
    m_haveNotify = false;
#ifndef WIN32
    QDir dir;
    QStringList paths = QString(getenv("PATH")).split(':');
    foreach(QString path, paths)
    {
        if(dir.exists(path+"/notify-send"))
        {
            m_haveNotify = true;
            break;
        }
    }
#endif // WIN32
}

dxutils::~dxutils()
{
#ifdef HAVE_ENCHANT
    Checkers::const_iterator it = m_checkers.begin();
    for(; it != m_checkers.end(); it++)
    {
        delete (*it).second;
    }
#endif
}

dxutils &dxutils::instance()
{
    static dxutils idxutils;
    return idxutils;
}

bool dxutils::isBadchar(char c)
{
    switch(c) {
        case ' ':
        case ',':
        case '\0':
        case '\02':
        case '\03':
        case '\017':
        case '\021':
        case '\026':
        case '\035':
        case '\037':
        case '\n':
        case '\r':
        case '<':
        case '>':
        case '"':
        case '\'':
            return true;
        default:
            return false;
    }
}

// checks is char nick/word delimiter
bool dxutils::isDelimiter(char c)
{
    switch(c) {
        case ' ':
        case '.':
        case ',':
        case '/':
        case '\\':
        case '`':
        case '\'':
        case '!':
        case '(':
        case ')':
        case '{':
        case '}':
        case '|':
        case '[':
        case ']':
        case '\"':
        case ':':
        case ';':
        case '<':
        case '>':
        case '?':
            return true;
        default:
            return false;
    }
}

//Return file size in human readable form
QString dxutils::getFileSize(quint64 size)
{
    double fsize = 0.0;
    if(size > 1000000000)
    {
        fsize = size/1073741824.0;
        return QString("%1 %2").arg(fsize, 0, 'f', 2).arg(_("GB"));
    }
    if(size > 100000)
    {
        fsize = size/1048576.0;
        return QString("%1 %2").arg(fsize, 0, 'f', 2).arg(_("MB"));
    }
    if(size > 1000)
    {
        fsize = size/1024.0;
        return QString("%1 %2").arg(fsize, 0, 'f', 2).arg(_("KB"));
    }
    return QString("%1 %2").arg(size).arg(_("bytes"));
}

QString dxutils::getFileSize(const QString& ssize)
{
    return getFileSize(ssize.toLongLong());
}

//std::string to QString
QString dxutils::fromStdString(const std::string &str)
{
    return QString::fromUtf8(str.c_str());
}

//QString to std::string
std::string dxutils::toStdString(const QString &str)
{
    return str.toUtf8().constData();
}

QString dxutils::getNickColor(const QString &nick)
{
    register int len = nick.length();
    register quint32 h = 0;
    for(register int i=0; i<len; i++)
        h = ((h << 5) + h) ^ nick[i].toLatin1();
    switch(h%8)
    {
        case 0: return QColor(196,160,0).name();
        case 1: return QColor(206,92,0).name();
        case 2: return QColor(143,89,2).name();
        case 3: return QColor(78,154,6).name();
        case 4: return QColor(32,74,135).name();
        case 5: return QColor(117,80,123).name();
        case 6: return QColor(164,0,0).name();
        default: return QColor(85,87,83).name();
    }
}

/*Return download/send speed in human readable form
 * speed = difference between two position during 1 second
 */
QString dxutils::getSpeed(quint64 speed)
{
    double fspeed = 0.0;
    if(speed > 100000)
    {
        fspeed = speed/(1048576.0);
        return _("%1 MB/s").arg(fspeed, 0, 'f', 2);
    }
    fspeed = speed/(1024.0);
    return _("%1 KB/s").arg(fspeed, 0, 'f', 2);
}

//Return remaining time of download/send
QString dxutils::getRemaining(quint64 size, quint64 speed)
{
    if(!speed) return "?";
    quint64 remainingTime = size/speed;
    if(remainingTime <= 0 || remainingTime > 54000) return "?";
    else
    {
        quint64 secs = remainingTime;
        quint64 hours = remainingTime/3600;
        secs -= hours*3600;
        quint64 mins = secs/60;
        secs -= mins*60;
        return QString("%1:%2:%3").arg(hours).arg(mins, 2, 10, QLatin1Char('0')).arg(secs, 2, 10, QLatin1Char('0'));
    }
}

void dxutils::showNotify(QString text, int position)
{
    if(m_haveNotify)
    {
        QString exec = QString("notify-send -i %1 dxirc '%2'").arg(dxutils::getDataPath()+"/icons/big_dxirc.png").arg(text);
        system(exec.toUtf8().constData());
    }
    else
    {
        text.replace("\n", "<br>");
        if(!m_firstNotify || position!=m_notifyPosition)
        {
            m_firstNotify = true;
            m_notifiesHeight = 0;
            m_notifyPosition = position;
        }
        NotifyWindow *notify = new NotifyWindow(text, m_notifiesHeight, 5);
        notify->showNotify(position);
        m_notifiesHeight += notify->height()+4;
    }
}

void dxutils::removeNotify()
{
    if(m_firstNotify)
        m_firstNotify = !m_firstNotify;
}

void dxutils::debugLine(const QString &line)
{
#ifdef DEBUG
    qDebug() << QTime::currentTime().toString("[hh:mm:ss] ") << line;
#endif
}

//is file with name on path right dxirc log file?
bool dxutils::isRightLogFile(const QString& path, const QString& name)
{
    if(QRegExp("^\\d\\d\\d\\d-\\d\\d-\\d\\d+$").exactMatch(name))
    {
        QFile file(path);
        if(!file.open(QIODevice::ReadOnly | QIODevice::Text))
            return false;
        char buf[22];
        qint64 lineLength = file.readLine(buf, sizeof(buf));
        if(lineLength != -1)
        {
            if(QRegExp("^\\[(.*)(\\d\\d:\\d\\d:\\d\\d\\])").indexIn(QString(buf)) != -1)
                return true;
            else
                return false;
        }
        else
            return false;
    }
    return false;
}

//get month string
QString dxutils::getMonth(int month)
{
    switch(month) {
        case 1: return _("January");
        case 2: return _("February");
        case 3: return _("March");
        case 4: return _("April");
        case 5: return _("May");
        case 6: return _("June");
        case 7: return _("July");
        case 8: return _("August");
        case 9: return _("September");
        case 10: return _("October");
        case 11: return _("November");
        case 12: return _("December");
        default: return "Rok ma jen 12 mesicu:)";
    }
}

//get month string-number for name
QString dxutils::getMonthNumber(const QString& month)
{
    if(month==_("January")) return "01";
    else if(month==_("February")) return "02";
    else if(month==_("March")) return "03";
    else if(month==_("April")) return "04";
    else if(month==_("May")) return "05";
    else if(month==_("June")) return "06";
    else if(month==_("July")) return "07";
    else if(month==_("August")) return "08";
    else if(month==_("September")) return "09";
    else if(month==_("October")) return "10";
    else if(month==_("November")) return "11";
    else if(month==_("December")) return "12";
    return "00";
}

void dxutils::playFile(const QString &file)
{
    QDir dir;
    if(dir.exists(file))
    {
#if(QT_VERSION < 0x050000)
#if defined(Q_OS_LINUX) || defined(Q_OS_UNIX)
        static const char * players[]={"aplay","play","esdplay","artsplay","ossplay",NULL};
        QStringList paths = QString(getenv("PATH")).split(':');
        foreach(QString path, paths)
        {
            for(int i=0; players[i]!=NULL; i++)
            {
                if(dir.exists(path+"/"+players[i]))
                {
                    QString exec = path+"/"+players[i];
                    if(exec.contains("aplay"))
                        exec.append(" -q");
                    exec += " "+file+" &";
                    system(exec.toLatin1().constData());
                }
            }
        }
#endif
#endif
        QSound::play(file);
    }
}

QString dxutils::createSmileys(QString text)
{
    for(QMap<QString,QString>::const_iterator it=preferences.m_smileysMap.constBegin(); it!=preferences.m_smileysMap.constEnd(); it++)
    {
        if(text.contains(it.key()))
        {
            text.replace(it.key(), "<img src='"+it.value()+"'>");
        }
    }
    return text;
}

//check path for nick icons
QString dxutils::checkThemePath(const QString &path)
{
    if(path == "internal") return path;
    QString themeDefaultPath = "internal";
    if(QFileInfo(path+"/irc_normal.png").exists()) return path;
    return themeDefaultPath;
}

//check path list for nick icons
QString dxutils::checkThemesList(const QString &list)
{
    QString themeDefaultPath = "internal;";
    QString themes;
    QStringList pathes = list.split(';', QString::SkipEmptyParts);
    for(int i=0; i<pathes.size(); ++i)
    {
        if(QFileInfo(pathes[i]+"/irc_normal.png").exists()) themes.append(pathes[i]+";");
    }
    if(!themes.contains(themeDefaultPath)) themes.prepend(themeDefaultPath);
    if(!themes.isEmpty()) return themes;
    return themeDefaultPath;
}

bool dxutils::haveNotify()
{
    return m_haveNotify;
}

QString dxutils::encrypt(const QString &text)
{
    QString result = "";
    for(int i=0; i<text.length(); i++)
    {
        result += text[i];
        int r = rand()%127;
        result += char(r<33 ? r+33 : r);
    }
    return result;
}

QString dxutils::decrypt(const QString &text)
{
    QString result = "";
    for(int i=0; i<text.length(); i++)
    {
        if((i+1)%2) result += text[i];
    }
    return result;
}

QString dxutils::getParam(QString toParse, int n, bool toEnd)
{
    return getParam(toParse, n, toEnd, ' ');
}

QString dxutils::getParam(QString toParse, int n, bool toEnd, const char &separator)
{
    if(toEnd)
    {
        return toParse.section(separator, n-1);
    }
    else
    {
        QString result = toParse.section(separator, n-1, n-1);
        if(result.isEmpty()) return toParse.section(separator, -1);
        else return result;
    }
}

QString dxutils::removeSpaces(const QString &text)
{
    if(text.contains(' '))
    {
        QString removed;
        for(int i=0; i<=text.length(); i++)
        {
            if(text[i].toLatin1() != 32)
                removed += text[i];
        }
        return removed;
    }
    else
        return text;
}

QString dxutils::removeNonalphanumeric(const QString &text)
{
    QString result;
    for(int i=0; i<text.length(); i++)
    {
        if((47<text[i].toLatin1()&&text[i].toLatin1()<58) ||
                (64<text[i].toLatin1()&&text[i].toLatin1()<91) ||
                (96<text[i].toLatin1()&&text[i].toLatin1()<123))
            result += text[i];
        else
            result += '0';
    }
    return result;
}

QString dxutils::createModes(char sign, char mode, QString nicks)
{
    QString modes;
    modes += sign;
    QString tomode;
    if(!nicks.contains(' ')) return modes+mode+" "+nicks;
    if(nicks.right(1) != " ") nicks.append(" ");
    while(nicks.contains(' '))
    {
        modes += mode;
        tomode += nicks.section(' ',0,0)+" ";
        nicks = nicks.section(' ',1);
    }
    return modes+" "+tomode;
}

//from kde source, thx
bool dxutils::isUtf8(const char *buf)
{
    int i, n;
    register unsigned char c;
    bool gotone = false;
    if (!buf)
    return true; // whatever, just don't crash

    #define F 0   /* character never appears in text */
    #define T 1   /* character appears in plain ASCII text */
    #define I 2   /* character appears in ISO-8859 text */
    #define X 3   /* character appears in non-ISO extended ASCII (Mac, IBM PC) */

    static const unsigned char text_chars[256] = {
        /*                  BEL BS HT LF    FF CR    */
        F, F, F, F, F, F, F, T, T, T, T, F, T, T, F, F,  /* 0x0X */
        /*                              ESC          */
        F, F, F, F, F, F, F, F, F, F, F, T, F, F, F, F,  /* 0x1X */
        T, T, T, T, T, T, T, T, T, T, T, T, T, T, T, T,  /* 0x2X */
        T, T, T, T, T, T, T, T, T, T, T, T, T, T, T, T,  /* 0x3X */
        T, T, T, T, T, T, T, T, T, T, T, T, T, T, T, T,  /* 0x4X */
        T, T, T, T, T, T, T, T, T, T, T, T, T, T, T, T,  /* 0x5X */
        T, T, T, T, T, T, T, T, T, T, T, T, T, T, T, T,  /* 0x6X */
        T, T, T, T, T, T, T, T, T, T, T, T, T, T, T, F,  /* 0x7X */
        /*            NEL                            */
        X, X, X, X, X, T, X, X, X, X, X, X, X, X, X, X,  /* 0x8X */
        X, X, X, X, X, X, X, X, X, X, X, X, X, X, X, X,  /* 0x9X */
        I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,  /* 0xaX */
        I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,  /* 0xbX */
        I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,  /* 0xcX */
        I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,  /* 0xdX */
        I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I,  /* 0xeX */
        I, I, I, I, I, I, I, I, I, I, I, I, I, I, I, I   /* 0xfX */
    };

    /* *ulen = 0; */
    for (i = 0; (c = buf[i]); i++)
    {
        if ((c & 0x80) == 0) /* 0xxxxxxx is plain ASCII */
        {
        /*
        * Even if the whole file is valid UTF-8 sequences,
        * still reject it if it uses weird control characters.
        */

        if (text_chars[c] != T)
            return false;

        }
        else if ((c & 0x40) == 0) /* 10xxxxxx never 1st byte */
        {
        return false;
        }
        else /* 11xxxxxx begins UTF-8 */
        {
            int following;

            if ((c & 0x20) == 0) {             /* 110xxxxx */
            following = 1;
            }
            else if ((c & 0x10) == 0) {      /* 1110xxxx */
            following = 2;
            }
            else if ((c & 0x08) == 0) {      /* 11110xxx */
            following = 3;
            }
            else if ((c & 0x04) == 0) {      /* 111110xx */
            following = 4;
            }
            else if ((c & 0x02) == 0) {      /* 1111110x */
            following = 5;
            }
            else
                return false;

            for (n = 0; n < following; n++) {
                i++;
                if (!(c = buf[i]))
                goto done;

                if ((c & 0x80) == 0 || (c & 0x40))
                return false;
            }
            gotone = true;
        }
    }
    done:
    return gotone;   /* don't claim it's UTF-8 if it's all 7-bit */
}

//get default spellchecking language
#ifdef HAVE_ENCHANT
QString dxutils::getDefaultLang()
{
    if(m_locale.isEmpty()) getLocale();
    if(m_checkers.find(m_locale) != m_checkers.end())
        return m_locale;
    Checkers::const_iterator it = m_checkers.begin();
    QString language = m_locale.section('_',0,0);
    for(; it != m_checkers.end(); it++)
    {
        if(QString::compare((*it).first.section('_',0,0),language)==0) return language;
    }
    if(m_availableSpellLangs.size()) return m_availableSpellLangs[0];
    else return "";
}
#else
QString dxutils::getDefaultLang()
{
    if(m_locale.isEmpty()) getLocale();
    return m_locale;
}
#endif

//return locale, e.g cs_CZ etc.
QString dxutils::getLocale()
{
    if(!m_locale.isEmpty())
        return m_locale;
#ifdef WIN32
    LCID lcid = GetUserDefaultLCID();
    if(lcid!=0)
    {
        quint32 localeid = LANGIDFROMLCID(lcid);
        /* locale code from this page
         * http://msdn.microsoft.com/en-us/library/dd318693%28v=VS.85%29.aspx
         * not fully implemented
         */
        switch(localeid){
            case 0x0436: m_locale = "af_ZA";break; // Afrikaans
            case 0x041c: m_locale = "sq_AL";break; // Albanian
            case 0x0401: m_locale = "ar_SA";break; // Arabic - Saudi Arabia
            case 0x0801: m_locale = "ar_IQ";break; // Arabic - Iraq
            case 0x0c01: m_locale = "ar_EG";break; // Arabic - Egypt
            case 0x1001: m_locale = "ar_LY";break; // Arabic - Libya
            case 0x1401: m_locale = "ar_DZ";break; // Arabic - Algeria
            case 0x1801: m_locale = "ar_MA";break; // Arabic - Morocco
            case 0x1c01: m_locale = "ar_TN";break; // Arabic - Tunisia
            case 0x2001: m_locale = "ar_OM";break; // Arabic - Oman
            case 0x2401: m_locale = "ar_YE";break; // Arabic - Yemen
            case 0x2801: m_locale = "ar_SY";break; // Arabic - Syria
            case 0x2c01: m_locale = "ar_JO";break; // Arabic - Jordan
            case 0x3001: m_locale = "ar_LB";break; // Arabic - Lebanon
            case 0x3401: m_locale = "ar_KW";break; // Arabic - Kuwait
            case 0x3801: m_locale = "ar_AE";break; // Arabic - United Arab Emirates
            case 0x3c01: m_locale = "ar_BH";break; // Arabic - Bahrain
            case 0x4001: m_locale = "ar_QA";break; // Arabic - Qatar
            case 0x042b: m_locale = "hy_AM";break; // Armenian
            case 0x042c: m_locale = "az_AZ";break; // Azeri Latin
            case 0x082c: m_locale = "az_AZ";break; // Azeri - Cyrillic
            case 0x042d: m_locale = "eu_ES";break; // Basque
            case 0x0423: m_locale = "be_BY";break; // Belarusian
            case 0x0445: m_locale = "bn_IN";break; // Begali
            case 0x201a: m_locale = "bs_BA";break; // Bosnian
            case 0x141a: m_locale = "bs_BA";break; // Bosnian - Cyrillic
            case 0x047e: m_locale = "br_FR";break; // Breton - France
            case 0x0402: m_locale = "bg_BG";break; // Bulgarian
            case 0x0403: m_locale = "ca_ES";break; // Catalan
            case 0x0004: m_locale = "zh_CHS";break; // Chinese - Simplified
            case 0x0404: m_locale = "zh_TW";break; // Chinese - Taiwan
            case 0x0804: m_locale = "zh_CN";break; // Chinese - PRC
            case 0x0c04: m_locale = "zh_HK";break; // Chinese - Hong Kong S.A.R.
            case 0x1004: m_locale = "zh_SG";break; // Chinese - Singapore
            case 0x1404: m_locale = "zh_MO";break; // Chinese - Macao S.A.R.
            case 0x7c04: m_locale = "zh_CHT";break; // Chinese - Traditional
            case 0x041a: m_locale = "hr_HR";break; // Croatian
            case 0x101a: m_locale = "hr_BA";break; // Croatian - Bosnia
            case 0x0405: m_locale = "cs_CZ";break; // Czech
            case 0x0406: m_locale = "da_DK";break; // Danish
            case 0x048c: m_locale = "gbz_AF";break; // Dari - Afghanistan
            case 0x0465: m_locale = "div_MV";break; // Divehi - Maldives
            case 0x0413: m_locale = "nl_NL";break; // Dutch - The Netherlands
            case 0x0813: m_locale = "nl_BE";break; // Dutch - Belgium
            case 0x0409: m_locale = "en_US";break; // English - United States
            case 0x0809: m_locale = "en_GB";break; // English - United Kingdom
            case 0x0c09: m_locale = "en_AU";break; // English - Australia
            case 0x1009: m_locale = "en_CA";break; // English - Canada
            case 0x1409: m_locale = "en_NZ";break; // English - New Zealand
            case 0x1809: m_locale = "en_IE";break; // English - Ireland
            case 0x1c09: m_locale = "en_ZA";break; // English - South Africa
            case 0x2009: m_locale = "en_JA";break; // English - Jamaica
            case 0x2409: m_locale = "en_CB";break; // English - Carribbean
            case 0x2809: m_locale = "en_BZ";break; // English - Belize
            case 0x2c09: m_locale = "en_TT";break; // English - Trinidad
            case 0x3009: m_locale = "en_ZW";break; // English - Zimbabwe
            case 0x3409: m_locale = "en_PH";break; // English - Phillippines
            case 0x0425: m_locale = "et_EE";break; // Estonian
            case 0x0438: m_locale = "fo_FO";break; // Faroese
            case 0x0464: m_locale = "fil_PH";break; // Filipino
            case 0x040b: m_locale = "fi_FI";break; // Finnish
            case 0x040c: m_locale = "fr_FR";break; // French - France
            case 0x080c: m_locale = "fr_BE";break; // French - Belgium
            case 0x0c0c: m_locale = "fr_CA";break; // French - Canada
            case 0x100c: m_locale = "fr_CH";break; // French - Switzerland
            case 0x140c: m_locale = "fr_LU";break; // French - Luxembourg
            case 0x180c: m_locale = "fr_MC";break; // French - Monaco
            case 0x0462: m_locale = "fy_NL";break; // Frisian - Netherlands
            case 0x0456: m_locale = "gl_ES";break; // Galician
            case 0x0437: m_locale = "ka_GE";break; // Georgian
            case 0x0407: m_locale = "de_DE";break; // German - Germany
            case 0x0807: m_locale = "de_CH";break; // German - Switzerland
            case 0x0c07: m_locale = "de_AT";break; // German - Austria
            case 0x1007: m_locale = "de_LU";break; // German - Luxembourg
            case 0x1407: m_locale = "de_LI";break; // German - Liechtenstein
            case 0x0408: m_locale = "el_GR";break; // Greek
            case 0x0447: m_locale = "gu_IN";break; // Gujarati
            case 0x040d: m_locale = "he_IL";break; // Hebrew
            case 0x0439: m_locale = "hi_IN";break; // Hindi
            case 0x040e: m_locale = "hu_HU";break; // Hungarian
            case 0x040f: m_locale = "is_IS";break; // Icelandic
            case 0x0421: m_locale = "id_ID";break; // Indonesian
            case 0x045d: m_locale = "iu_CA";break; // Inuktitut
            case 0x085d: m_locale = "iu_CA";break; // Inuktitut - Latin
            case 0x083c: m_locale = "ga_IE";break; // Irish - Ireland
            case 0x0434: m_locale = "xh_ZA";break; // Xhosa - South Africa
            case 0x0435: m_locale = "zu_ZA";break; // Zulu
            case 0x0410: m_locale = "it_IT";break; // Italian - Italy
            case 0x0810: m_locale = "it_CH";break; // Italian - Switzerland
            case 0x0411: m_locale = "ja_JP";break; // Japanese
            case 0x044b: m_locale = "kn_IN";break; // Kannada - India
            case 0x043f: m_locale = "kk_KZ";break; // Kazakh
            case 0x0457: m_locale = "kok_IN";break; // Konkani
            case 0x0412: m_locale = "ko_KR";break; // Korean
            case 0x0440: m_locale = "ky_KG";break; // Kyrgyz
            case 0x0426: m_locale = "lv_LV";break; // Latvian
            case 0x0427: m_locale = "lt_LT";break; // Lithuanian
            case 0x046e: m_locale = "lb_LU";break; // Luxembourgish
            case 0x042f: m_locale = "mk_MK";break; // FYRO Macedonian
            case 0x043e: m_locale = "ms_MY";break; // Malay - Malaysia
            case 0x083e: m_locale = "ms_BN";break; // Malay - Brunei
            case 0x044c: m_locale = "ml_IN";break; // Malayalam - India
            case 0x043a: m_locale = "mt_MT";break; // Maltese
            case 0x0481: m_locale = "mi_NZ";break; // Maori
            case 0x047a: m_locale = "arn_CL";break; // Mapudungun
            case 0x044e: m_locale = "mr_IN";break; // Marathi
            case 0x047c: m_locale = "moh_CA";break; // Mohawk - Canada
            case 0x0450: m_locale = "mn_MN";break; // Mongolian
            case 0x0461: m_locale = "ne_NP";break; // Nepali
            case 0x0414: m_locale = "nb_NO";break; // Norwegian - Bokmal
            case 0x0814: m_locale = "nn_NO";break; // Norwegian - Nynorsk
            case 0x0482: m_locale = "oc_FR";break; // Occitan - France
            case 0x0448: m_locale = "or_IN";break; // Oriya - India
            case 0x0463: m_locale = "ps_AF";break; // Pashto - Afghanistan
            case 0x0429: m_locale = "fa_IR";break; // Persian
            case 0x0415: m_locale = "pl_PL";break; // Polish
            case 0x0416: m_locale = "pt_BR";break; // Portuguese - Brazil
            case 0x0816: m_locale = "pt_PT";break; // Portuguese - Portugal
            case 0x0446: m_locale = "pa_IN";break; // Punjabi
            case 0x046b: m_locale = "quz_BO";break; // Quechua (Bolivia)
            case 0x086b: m_locale = "quz_EC";break; // Quechua (Ecuador)
            case 0x0c6b: m_locale = "quz_PE";break; // Quechua (Peru)
            case 0x0418: m_locale = "ro_RO";break; // Romanian - Romania
            case 0x0417: m_locale = "rm_CH";break; // Raeto-Romanese
            case 0x0419: m_locale = "ru_RU";break; // Russian
            case 0x243b: m_locale = "smn_FI";break; // Sami Finland
            case 0x103b: m_locale = "smj_NO";break; // Sami Norway
            case 0x143b: m_locale = "smj_SE";break; // Sami Sweden
            case 0x043b: m_locale = "se_NO";break; // Sami Northern Norway
            case 0x083b: m_locale = "se_SE";break; // Sami Northern Sweden
            case 0x0c3b: m_locale = "se_FI";break; // Sami Northern Finland
            case 0x203b: m_locale = "sms_FI";break; // Sami Skolt
            case 0x183b: m_locale = "sma_NO";break; // Sami Southern Norway
            case 0x1c3b: m_locale = "sma_SE";break; // Sami Southern Sweden
            case 0x044f: m_locale = "sa_IN";break; // Sanskrit
            case 0x0c1a: m_locale = "sr_SP";break; // Serbian - Cyrillic
            case 0x1c1a: m_locale = "sr_BA";break; // Serbian - Bosnia Cyrillic
            case 0x081a: m_locale = "sr_SP";break; // Serbian - Latin
            case 0x181a: m_locale = "sr_BA";break; // Serbian - Bosnia Latin
            case 0x046c: m_locale = "ns_ZA";break; // Northern Sotho
            case 0x0432: m_locale = "tn_ZA";break; // Setswana - Southern Africa
            case 0x041b: m_locale = "sk_SK";break; // Slovak
            case 0x0424: m_locale = "sl_SI";break; // Slovenian
            case 0x040a: m_locale = "es_ES";break; // Spanish - Spain
            case 0x080a: m_locale = "es_MX";break; // Spanish - Mexico
            case 0x0c0a: m_locale = "es_ES";break; // Spanish - Spain (Modern)
            case 0x100a: m_locale = "es_GT";break; // Spanish - Guatemala
            case 0x140a: m_locale = "es_CR";break; // Spanish - Costa Rica
            case 0x180a: m_locale = "es_PA";break; // Spanish - Panama
            case 0x1c0a: m_locale = "es_DO";break; // Spanish - Dominican Republic
            case 0x200a: m_locale = "es_VE";break; // Spanish - Venezuela
            case 0x240a: m_locale = "es_CO";break; // Spanish - Colombia
            case 0x280a: m_locale = "es_PE";break; // Spanish - Peru
            case 0x2c0a: m_locale = "es_AR";break; // Spanish - Argentina
            case 0x300a: m_locale = "es_EC";break; // Spanish - Ecuador
            case 0x340a: m_locale = "es_CL";break; // Spanish - Chile
            case 0x380a: m_locale = "es_UR";break; // Spanish - Uruguay
            case 0x3c0a: m_locale = "es_PY";break; // Spanish - Paraguay
            case 0x400a: m_locale = "es_BO";break; // Spanish - Bolivia
            case 0x440a: m_locale = "es_SV";break; // Spanish - El Salvador
            case 0x480a: m_locale = "es_HN";break; // Spanish - Honduras
            case 0x4c0a: m_locale = "es_NI";break; // Spanish - Nicaragua
            case 0x500a: m_locale = "es_PR";break; // Spanish - Puerto Rico
            case 0x0441: m_locale = "sw_KE";break; // Swahili
            case 0x041d: m_locale = "sv_SE";break; // Swedish - Sweden
            case 0x081d: m_locale = "sv_FI";break; // Swedish - Finland
            case 0x045a: m_locale = "syr_SY";break; // Syriac
            case 0x0449: m_locale = "ta_IN";break; // Tamil
            case 0x0444: m_locale = "tt_RU";break; // Tatar
            case 0x044a: m_locale = "te_IN";break; // Telugu
            case 0x041e: m_locale = "th_TH";break; // Thai
            case 0x041f: m_locale = "tr_TR";break; // Turkish
            case 0x0422: m_locale = "uk_UA";break; // Ukrainian
            case 0x0420: m_locale = "ur_PK";break; // Urdu
            case 0x0820: m_locale = "ur_IN";break; // Urdu - India
            case 0x0443: m_locale = "uz_UZ";break; // Uzbek - Latin
            case 0x0843: m_locale = "uz_UZ";break; // Uzbek - Cyrillic
            case 0x042a: m_locale = "vi_VN";break; // Vietnamese
            case 0x0452: m_locale = "cy_GB";break; // Welsh
            default: m_locale = "en_US";
        }
    }
    else m_locale = "en_US";
#else
    if(!QProcessEnvironment::systemEnvironment().value("LANG").isEmpty()) m_locale = QProcessEnvironment::systemEnvironment().value("LANG");
    else if(!QProcessEnvironment::systemEnvironment().value("LC_ALL").isEmpty()) m_locale = QProcessEnvironment::systemEnvironment().value("LC_ALL");
    else if(!QProcessEnvironment::systemEnvironment().value("LC_MESSAGES").isEmpty()) m_locale = QProcessEnvironment::systemEnvironment().value("LC_MESSAGES");
    else m_locale = "en_US";
    if(m_locale == "C" || m_locale == "POSIX")
    {
        m_locale = "en_US";
    }
    m_locale = m_locale.section('.',0,0).section('@',0,0);
#endif
    return m_locale;
}

void dxutils::loadAlias()
{
    QSettings set(preferences.getIniFile(), QSettings::IniFormat);
    set.beginGroup("ALIASES");
    int num = set.value("number", 0).toInt();
    m_aliases.clear();
    if(num)
    {
        for(int i=0; i<num; i++)
        {
            QString key, value;
            key = set.value(QString("key%1").arg(i), "").toString();
            value = set.value(QString("value%1").arg(i), "").toString();
            if(!key.isEmpty() && !value.isEmpty())
                m_aliases.insert(key, value);
        }
    }
    set.endGroup();
    fillCommands();
}

void dxutils::setAliases(QMap<QString,QString> a)
{
    m_aliases = a;
    fillCommands();
}

QString dxutils::getAlias(QString key)
{
    if(m_aliases.count(key.toLower())>0)
        return m_aliases.find(key).value();
    return "";
}

QMap<QString,QString> dxutils::getAliases()
{
    return m_aliases;
}

void dxutils::fillCommands()
{
    m_commands.clear();
    m_dcccommands.clear();
    m_commands.append("ADMIN");
    m_commands.append("AWAY");
    m_commands.append("BANLIST");
    m_commands.append("BOATS");
    m_commands.append("CONNECT");
    m_commands.append("COMMANDS");
    m_dcccommands.append("COMMANDS");
    m_commands.append("CTCP");
    m_commands.append("CYCLE");
    m_commands.append("DCC");
    m_commands.append("DEOP");
    m_commands.append("DEVOICE");
    m_commands.append("DISCONNECT");
    m_commands.append("EXEC");
    m_dcccommands.append("EXEC");
    m_commands.append("HELP");
    m_dcccommands.append("HELP");
    m_commands.append("IGNORE");
    m_commands.append("INVITE");
    m_commands.append("JOIN");
    m_commands.append("KICK");
    m_commands.append("KILL");
    m_commands.append("LIST");
    m_commands.append("LOG");
    m_dcccommands.append("LOG");
#ifdef HAVE_LUA
    m_commands.append("LUA");
    m_dcccommands.append("LUA");
#endif
    m_commands.append("ME");
    m_dcccommands.append("ME");
    m_commands.append("MODE");
    m_commands.append("MSG");
    m_commands.append("NAMES");
    m_commands.append("NICK");
    m_commands.append("NOTICE");
    m_commands.append("OP");
    m_commands.append("OPER");
    m_commands.append("PART");
    m_commands.append("QUERY");
    m_commands.append("QUIT");
    m_commands.append("QUOTE");
    m_commands.append("SAY");
    m_dcccommands.append("SAY");
    m_commands.append("STATS");
    m_commands.append("TIME");
    m_commands.append("TOPIC");
    m_commands.append("VOICE");
    m_commands.append("WALLOPS");
    m_commands.append("WHO");
    m_commands.append("WHOAMI");
    m_commands.append("WHOIS");
    m_commands.append("WHOWAS");
    QMap<QString,QString>::iterator it;
    for(it=m_aliases.begin(); it!=m_aliases.end(); it++)
    {
        m_commands.append(it.key().section('/',1));
    }
    for(int i=0; i<m_scriptCommands.size(); i++)
    {
        m_commands.append(m_scriptCommands[i].name);
        m_dcccommands.append(m_scriptCommands[i].name);
    }
    qSort(m_commands.begin(), m_commands.end(), caseInsensitiveLessThan);
    qSort(m_dcccommands.begin(), m_dcccommands.end(), caseInsensitiveLessThan);
}

int dxutils::commandsNo()
{
    return m_commands.size();
}

QString dxutils::commandsAt(int i)
{
    return m_commands.at(i);
}

bool dxutils::isCommand(const QString &command)
{
    if(!m_commands.size()) return false;
    for(int i=0; i<m_commands.size(); i++)
    {
        if(QString::compare(command, m_commands.at(i)) == 0) return true;
    }
    return false;
}

int dxutils::dcccommandsNo()
{
    return m_dcccommands.size();
}

QString dxutils::dcccommandsAt(int i)
{
    return m_dcccommands.at(i);
}

bool dxutils::isDccCommand(const QString &command)
{
    if(!m_dcccommands.size()) return false;
    for(int i=0; i<m_dcccommands.size(); i++)
    {
        if(QString::compare(command, m_dcccommands.at(i)) == 0) return true;
    }
    return false;
}

void dxutils::addScriptCommand(LuaScriptCommand command)
{
    m_scriptCommands.append(command);
    fillCommands();
}

//Remove one command for lua script
bool dxutils::removeScriptCommand(const QString &command)
{
    for(int i=m_scriptCommands.size()-1; i>-1; i--)
    {
        if(QString::compare(command, m_scriptCommands[i].name) == 0)
        {
            m_scriptCommands.removeAt(i);
            fillCommands();
            return true;
        }
    }
    return false;
}

//Remove all commands for lua script
bool dxutils::removeScriptCommands(const QString& script)
{
    bool result = false;
    for(int i=m_scriptCommands.size()-1; i>-1; i--)
    {
        if(QString::compare(script, m_scriptCommands[i].script) == 0)
        {
            m_scriptCommands.removeAt(i);
            result = true;
        }
    }
    fillCommands();
    return result;
}

bool dxutils::isScriptCommand(const QString &command)
{
    for(int i=0; i<m_scriptCommands.size(); i++)
    {
        if(QString::compare(command, m_scriptCommands[i].name) == 0) return true;
    }
    return false;
}

QString dxutils::availableCommands()
{
    QString commandstr = _("Available commands: ");
    for(int i=0; i < m_commands.size(); i++)
    {
        if(m_commands[i] != "commands") commandstr += m_commands[i].toUpper()+(i != m_commands.size() - 1? ", " : "");
    }
    return commandstr;
}

QString dxutils::availableDccCommands()
{
    QString commandstr = _("Available commands: ");
    for(int i=0; i < m_dcccommands.size(); i++)
    {
        if(m_dcccommands[i] != "commands") commandstr += m_dcccommands[i].toUpper()+(i != m_dcccommands.size() - 1? ", " : "");
    }
    return commandstr;
}

QString dxutils::availableScriptsCommands()
{
    QString commandstr = _("Available commands: ");
    for(int i=0; i < m_scriptCommands.size(); i++)
    {
        commandstr += m_scriptCommands[i].name.toUpper()+(i != m_scriptCommands.size() - 1? ", " : "");
    }
    return commandstr;
}

QString dxutils::availableScriptCommands(const QString& script)
{
    bool result = false;
    QString commandstr = _("Script commands: ");
    for(int i=m_scriptCommands.size()-1; i>-1; i--)
    {
        if(QString::compare(script, m_scriptCommands[i].script) == 0)
        {
            if(result) commandstr += ", ";
            commandstr += m_scriptCommands[i].name.toUpper();
            result = true;
        }
    }
    if(result) return commandstr;
    else return _("Script hasn't commands");
}

QString dxutils::getHelpText(const QString &command)
{
    for(int i=0; i<m_scriptCommands.size(); i++)
    {
        if(QString::compare(command, m_scriptCommands[i].name) == 0) return m_scriptCommands[i].helptext;
    }
    return _("Command %1 doesn't exists").arg(command);
}

QString dxutils::getFuncname(const QString &command)
{
    for(int i=0; i<m_scriptCommands.size(); i++)
    {
        if(QString::compare(command, m_scriptCommands[i].name) == 0) return m_scriptCommands[i].funcname;
    }
    return _("Command %1 doesn't exists").arg(command);
}

QString dxutils::getScriptName(const QString &command)
{
    for(int i=0; i<m_scriptCommands.size(); i++)
    {
        if(QString::compare(command, m_scriptCommands[i].name) == 0) return m_scriptCommands[i].script;
    }
    return _("Command %1 doesn't exists").arg(command);
}

QString dxutils::getStringIniEntry(const QString &section, const QString &key, const QString &def)
{
    QSettings set(preferences.getIniFile(), QSettings::IniFormat);
    set.beginGroup(section);
    return set.value(key, def).toString();
}

int dxutils::getIntIniEntry(const QString &section, const QString &key, int def)
{
    QSettings set(preferences.getIniFile(), QSettings::IniFormat);
    set.beginGroup(section);
    return set.value(key, def).toInt();
}

//return bool entry in inifile
bool dxutils::getBoolIniEntry(const QString &section, const QString &key, bool def)
{
    QSettings set(preferences.getIniFile(), QSettings::IniFormat);
    set.beginGroup(section);
    return set.value(key, def).toBool();
}

// return color entry in inifile
QColor dxutils::getColorIniEntry(const QString &section, const QString &key, QColor def)
{
    QSettings set(preferences.getIniFile(), QSettings::IniFormat);
    set.beginGroup(section);
    return set.value(key, def).value<QColor>();
}

//true if word right
#ifdef HAVE_ENCHANT
bool dxutils::checkWord(QString word, QString lang)
{
    Checkers::const_iterator it = m_checkers.find(lang);
    if(it != m_checkers.end())
        return (*it).second->check(toStdString(word));
    return true;
}
#else
#ifdef Q_OS_DARWIN
bool dxutils::checkWord(QString word, QString)
{
    if(_instance) _instance = new MacSpellCheck;
    return _instance->checkWord(word);
}
#else
bool dxutils::checkWord(QString, QString)
{
    return true;
}
#endif //Q_OS_DARWIN
#endif //HAVE_ENCHANT

//fill available langs
#ifdef HAVE_ENCHANT
void dxutils::loadAvailableSpellLangs()
{
    CheckerPair chpair(&m_checkers, &m_availableSpellLangs);
    enchant::Broker::instance()->list_dicts(enumerateDicts, &chpair);
    if(m_availableSpellLangs.size())
    {
        m_availableSpellLangs.sort();
    }
    for(int i=0; i<m_availableSpellLangs.size(); i++)
    {
        if(m_checkers.find(m_availableSpellLangs[i]) != m_checkers.end())
            return;
        try
        {
            m_checkers[m_availableSpellLangs[i]] = enchant::Broker::instance()->request_dict(toStdString(m_availableSpellLangs[i]));
        }
        catch(enchant::Exception &e)
        {
        }
    }
}
#else
void dxutils::loadAvailableSpellLangs()
{
    m_availableSpellLangs.clear();
}
#endif

//get number of spellchecking language
#ifdef HAVE_ENCHANT
int dxutils::getAvailableSpellLangsNum()
{
    return m_availableSpellLangs.size();
}
#else
int dxutils::getAvailableSpellLangsNum()
{
    return 0;
}
#endif

//return array of spellchecking language
QStringList dxutils::getAvailableSpellLangs()
{
    return m_availableSpellLangs;
}

//time stamp to human readable string
QString dxutils::stamp2string(QString datestr, QString timestr)
{
    return QString("["+QDate(datestr.section('-',0,0).toInt(), datestr.section('-',1,1).toInt(), datestr.section('-',2).toInt()).toString(Qt::SystemLocaleShortDate)+" "+timestr+"]");
}

//normalize log line
QString dxutils::normalizeLogLine(QString datestr, std::string line)
{
    QString nline = fromStdString(line);
    QString date = stamp2string(datestr, nline.section(']',0,0).section('[',1));
    nline = date+nline.section(']',1);
    return nline;
}

//Remove ircmark from text
QString dxutils::stripColors(const QString &text, const bool stripOther)
{
    QString newstr;
    bool color = false;
    int numbers = 0;
    for(int i=0; i < text.length(); i++) {
        if(text[i] == '\017') //reset
        {
            color = false;
        }
        else if(stripOther && text[i] == '\002')
        {
            //remove bold mark
        }
        else if(stripOther && text[i] == '\037')
        {
            //remove underline mark
        }
        else if(text[i] == '\035')
        {
            //remove italic mark
        }
        else if(text[i] == '\021')
        {
            //remove fixed mark
        }
        else if(text[i] == '\026')
        {
            //remove reverse mark
        }
        else if(text[i] == '\003') //color
        {
            color = true;
        }
        else if(color && isdigit(text[i].toLatin1()) && numbers < 2)
        {
            numbers++;
        }
        else if(color && text[i] == ',' && numbers < 3)
        {
            numbers = 0;
        }
        else
        {
            numbers = 0;
            color = false;
            newstr += text[i];
        }
    }
    return newstr;
}

//Return color for ircolor code
QColor dxutils::getIrcColor(int code)
{
    switch(code){
        case 0:
            return QColor("white");
        case 1:
            return QColor("black");
        case 2:
            return QColor(0,0,128); //blue
        case 3:
            return QColor(0,128,0); //green
        case 4:
            return QColor(255,0,0); //lightred
        case 5:
            return QColor(128,0,64); //brown
        case 6:
            return QColor(128,0,128); //purple
        case 7:
            return QColor(255,128,64); //orange
        case 8:
            return QColor(255,255,0); //yellow
        case 9:
            return QColor(128,255,0); //lightgreen
        case 10:
            return QColor(0,128,128); //cyan
        case 11:
            return QColor(0,255,255); //lightcyan
        case 12:
            return QColor(0,0,255); //lightblue
        case 13:
            return QColor(255,0,255); //pink
        case 14:
            return QColor(128,128,128); //grey
        case 15:
            return QColor(192,192,192); //lightgrey
        default:
            return preferences.m_colors.text;
    }
}

//Remove ircmark from text
QString dxutils::stripLogMark(const QString &text)
{
    QString newstr;
    bool mark = false;
    int numbers = 0;
    int i = 0;
    while(text[i] != '\0') {
        if(text[i] == '\014') //log mark
        {
            if(!mark) mark = true; //first mark
            else mark = false; //last mark
        }
        else if(mark && isdigit(text[i].toLatin1()) && numbers < 2)
        {
            numbers++;
        }
        else
        {
            numbers = 0;
            mark = false;
            newstr += text[i];
        }
        i++;
    }
    return newstr;
}

QColor dxutils::getColorForStyle(int style)
{
    switch (style) {
        //gray text - user commands
        case 1: return preferences.m_colors.user;
        //orange text - Actions
        case 2: return preferences.m_colors.action;
        //blue text - Notice
        case 3: return preferences.m_colors.notice;
        //red text - Errors
        case 4: return preferences.m_colors.error;
        //highlight text
        case 8: return preferences.m_colors.hilight;
        //mymsg style
        case 10: return preferences.m_colors.mymsg;
        //log line style
        case 11: return preferences.m_colors.log;
        default: return preferences.m_colors.text;
    }
}

//get spellchecking language for channel
#ifdef HAVE_ENCHANT
QString dxutils::getChannelLang(const QString &channel)
{
    if(channel.isEmpty()) return getDefaultLang();
    QMap<QString,QString>::iterator it = preferences.m_langs.find(channel);
    QString lang = "";
    if(it != preferences.m_langs.end())
        lang = it.value();
    if(m_checkers.find(lang) != m_checkers.end())
        return lang;
    else return getDefaultLang();
}
#else
QString dxutils::getChannelLang(const QString&)
{
    return "";
}
#endif

QString dxutils::createLinks(const QString &txt, bool removeThan)
{
    QString line = "";
    int i = 0;
    int linkLength = 0;
    int length = txt.length();
    while(i<length)
    {
        if(txt[i]=='<' && removeThan)
        {
            line.append("&lt;");
        }
        else if(txt[i]=='>' && removeThan)
        {
            line.append("&gt;");
        }
        else if(txt[i]=='h' && !QString::compare(txt.mid(i,7),"http://",Qt::CaseInsensitive))
        {
            for(int j=i; j<length; j++)
            {
                if(isBadchar(txt[j].toLatin1()))
                {
                    break;
                }
                linkLength++;
            }
            if(linkLength>7)
                line.append(QString("<a href=\"%1\">%1</a>").arg(txt.mid(i, linkLength)));
            else
                line.append(txt.mid(i, linkLength));
            i+=linkLength-1;
            linkLength=0;
        }
        else if(txt[i]=='h' && !QString::compare(txt.mid(i,8),"https://",Qt::CaseInsensitive))
        {
            for(int j=i; j<length; j++)
            {
                if(isBadchar(txt[j].toLatin1()))
                {
                    break;
                }
                linkLength++;
            }
            if(linkLength>8)
                line.append(QString("<a href=\"%1\">%1</a>").arg(txt.mid(i, linkLength)));
            else
                line.append(txt.mid(i, linkLength));
            i+=linkLength-1;
            linkLength=0;
        }
        else if(txt[i]=='f' && !QString::compare(txt.mid(i,6),"ftp://",Qt::CaseInsensitive))
        {
            for(int j=i; j<length; j++)
            {
                if(isBadchar(txt[j].toLatin1()))
                {
                    break;
                }
                linkLength++;
            }
            if(linkLength>6)
                line.append(QString("<a href=\"%1\">%1</a>").arg(txt.mid(i, linkLength)));
            else
                line.append(txt.mid(i, linkLength));
            i+=linkLength-1;
            linkLength=0;
        }
        else if(txt[i]=='w' && !QString::compare(txt.mid(i,4),"www.",Qt::CaseInsensitive))
        {
            for(int j=i; j<length; j++)
            {
                if(isBadchar(txt[j].toLatin1()))
                {
                    break;
                }
                linkLength++;
            }
            if(linkLength>4)
                line.append(QString("<a href=\"%1\">%1</a>").arg(txt.mid(i, linkLength)));
            else
                line.append(txt.mid(i, linkLength));
            i+=linkLength-1;
            linkLength=0;
        }
        else
        {
            line.append(txt[i]);
        }
        i++;
    }
    return line;
}

static QString closeTags(QList<QString> tags)
{
    QString result;
    for(int i=tags.size()-1; i>-1; i--)
    {
        result.append("</");
        if(tags[i].contains(' ')) result.append(tags[i].section('<',1).section(' ',0,0));
        else result.append(tags[i].section('<',1).section('>',0,0));
        result.append('>');
    }
    return result;
}

static QList<QString> openTags(bool color, QColor foreColor, bool bgcolor, QColor backColor, bool bold, bool under, bool italic)
{
    QList<QString> result;
    if(color) result.append(QString("<font color=\"%1\">").arg(foreColor.name()));
    if(bgcolor) result.append(QString("<span style=\"background-color:%1\">").arg(backColor.name()));
    if(bold) result.append("<b>");
    if(under) result.append("<u>");
    if(italic) result.append("<i>");
    return result;
}

//removeThan is remove < and >
QString dxutils::parseText(const QString &txt, int stylenum, bool removeThan)
{
    const QString linked = createLinks(txt, removeThan);
    QString line = "";
    if(stylenum == 5) //bold text
    {
        line.prepend("<b>");
    }
    if(stylenum == 6) //underline style
    {
        line.prepend("<u>");
    }
    if(stylenum == 7) //bold & underline
    {
        line.prepend("<b><u>");
    }
    int i = 0;
    bool bold = false;
    bool under = false;
    bool italic = false;
    bool color = false;
    bool bgcolor = false;
    int length = linked.length();
    QList<QString> tags;
    QColor foreColor = dxutils::getColorForStyle(stylenum);
    QColor backColor = preferences.m_colors.back;
    if(dxutils::getColorForStyle(stylenum)!=preferences.m_colors.text) line.prepend(QString("<font color=\"%1\">").arg(foreColor.name()));
    while(i<length)
    {
        if(linked[i] == '\002') //bold
        {
            if(!tags.empty())
            {
                line.append(closeTags(tags));
            }
            bold = !bold;
            tags = openTags(color, foreColor, bgcolor, backColor, bold, under, italic);
            foreach(QString tag, tags)
                line.append(tag);
        }
        else if(linked[i] == '\026') //reverse
        {
            if(!tags.empty())
            {
                line.append(closeTags(tags));
            }
            bgcolor = true;
            color = true;
            QColor tempColor = foreColor;
            foreColor = backColor;
            backColor = tempColor;
            tags = openTags(color, foreColor, bgcolor, backColor, bold, under, italic);
            foreach(QString tag, tags)
                line.append(tag);
        }
        else if(linked[i] == '\037') //underline
        {
            if(!tags.empty())
            {
                line.append(closeTags(tags));
            }
            under = !under;
            tags = openTags(color, foreColor, bgcolor, backColor, bold, under, italic);
            foreach(QString tag, tags)
                line.append(tag);
        }
        else if(linked[i] == '\021') //fixed
        {
            dxutils::debugLine("Poslan fixed styl");
        }
        else if(linked[i] == '\035') //italic
        {
            if(!tags.empty())
            {
                line.append(closeTags(tags));
            }
            italic = !italic;
            tags = openTags(color, foreColor, bgcolor, backColor, bold, under, italic);
            foreach(QString tag, tags)
                line.append(tag);
        }
        else if(linked[i] == '\003') //color
        {
            if(!tags.empty())
            {
                line.append(closeTags(tags));
            }
            int colorLength = 0;
            foreColor = preferences.m_colors.text;
            backColor = preferences.m_colors.back;
            bool isHexColor = false;
            if(i+1<length)
            {
                if(linked[i+1] == '#') isHexColor = true;
            }
            if(isHexColor)
            {
                if(QRegExp("[0-9a-fA-F]{6}").exactMatch(linked.mid(i+2,6)))
                {
                    color = true;
                    foreColor = QColor(linked.mid(i+1,7));
                    colorLength +=7;
                }
                if(i+8 < length && linked[i+8] == ',' && QRegExp("[0-9a-fA-F]{6}").exactMatch(linked.mid(i+10,6)))
                {
                    bgcolor = true;
                    backColor = QColor(linked.mid(i+9,7));
                    colorLength +=8;
                }
            }
            else
            {
                if(i+2<length)
                {
                    int code = -1;
                    if(isdigit(linked[i+1].toLatin1()))
                    {
                        if(isdigit(linked[i+2].toLatin1()))
                        {
                            code = (linked[i+1].toLatin1()-48)*10+linked[i+2].toLatin1()-48;
                            colorLength +=2;
                        }
                        else
                        {
                            code = linked[i+1].toLatin1()-48;
                            colorLength ++;
                        }
                    }
                    if(code!=-1)
                    {
                        color = true;
                        foreColor = dxutils::getIrcColor(code%16);
                    }
                }
                if(i+colorLength+1 < length && linked[i+colorLength+1] == ',')
                {
                    int code = -1;
                    if(isdigit(linked[i+colorLength+2].toLatin1()))
                    {
                        if(isdigit(linked[i+colorLength+3].toLatin1()))
                        {
                            code = (linked[i+colorLength+2].toLatin1()-48)*10+linked[i+colorLength+3].toLatin1()-48;
                            colorLength +=3;
                        }
                        else
                        {
                            code = linked[i+colorLength+2].toLatin1()-48;
                            colorLength +=2;
                        }
                    }
                    if(code!=-1)
                    {
                        bgcolor = true;
                        backColor = dxutils::getIrcColor(code%16);
                    }
                }
            }
            tags = openTags(color, foreColor, bgcolor, backColor, bold, under, italic);
            foreach(QString tag, tags)
                line.append(tag);
            i +=colorLength;
        }
        else if(linked[i] == '\017') //reset
        {
            if(!tags.empty())
            {
                line.append(closeTags(tags));
            }
            bold = false;
            under = false;
            italic = false;
            bgcolor = false;
            color = false;
            foreColor = preferences.m_colors.text;
            backColor = preferences.m_colors.back;
        }
        else
        {
            line.append(linked[i]);
        }
        i++;
    }
    if(!tags.empty())
    {
        line.append(closeTags(tags));
    }
    if(stylenum == 5) //bold text
    {
        line.append("</b>");
    }
    if(stylenum == 6) //underline style
    {
        line.append("</u>");
    }
    if(stylenum == 7) //bold & underline
    {
        line.append("</u></b>");
    }
    if(dxutils::getColorForStyle(stylenum)!=preferences.m_colors.text) line.append("</font>");
    return createSmileys(line);
}

//logstyle = true for IrcTabItem::loadLogLines()
QString dxutils::parseLogline(const QString& line, bool logstyle)
{
    if(line.section(']',1)[1].toLatin1()=='\014') //log style mark
    {
        int style = logstyle?11:line.section('\014',1).section('\014',0,0).toInt();
        QString time = line.section('\014',0,0);
        if(line.section('\014',2)[0].toLatin1()=='<') //has nick
        {
            QString nick = line.section('<',1).section('>',0,0);
            QString msg = line.section('>',1);
            if(logstyle)
            {
                QString nickStyled = "";
                if(preferences.m_coloredNick) nickStyled = QString("<font color=\"%1\">%2: </font>").arg(getNickColor(nick)).arg(nick);
                else nickStyled = QString("<b>%1: </b>").arg(nick);
                return parseText(time+nickStyled+msg, style, false);
            }
            else
            {
                return time+parseText(nick+":"+msg, style);
            }
        }
        else
        {
            if(logstyle) return parseText(time+line.section('\014',2), style, false);
            else return time+parseText(line.section('\014',2), style);
        }
    }
    else
    {
        if(line.section(']',1)[1].toLatin1()=='<') //line with nick
        {
            QString time = line.section('<',0,0);
            QString nick = line.section('<',1).section('>',0,0);
            QString nickStyled = "";
            if(preferences.m_coloredNick) nickStyled = QString("<font color=\"%1\">%2: </font>").arg(getNickColor(nick)).arg(nick);
            else nickStyled = QString("<b>%1: </b>").arg(nick);
            QString msg = line.section('>',1).trimmed();
            if(logstyle) return parseText(time+nickStyled+msg, 11, false);
            else return time+nickStyled+parseText(msg, 0);
        }
        else
        {
            if(logstyle) return parseText(line, 11, false);
            else return parseText(line, 0);
        }
    }
}

//Get away color
QColor dxutils::makeAwayColor(QColor clr)
{
    int r,g,b;
    r=clr.red();
    g=clr.green();
    b=clr.blue();
    if(r==g && r==b && !r)
    {
        r=255;
        g=255;
        b=255;
    }
    r=(50*r)/100;
    g=(50*g)/100;
    b=(50*b)/100;
    return QColor(r,g,b);
}

//UserMode to QIcon
QString dxutils::modeToIcon(UserMode mode)
{
    switch(mode) {
    case OWNER:
    {
        return preferences.m_themePath+"/irc_owner.png";
    }
    case ADMIN:
    {
        return preferences.m_themePath+"/irc_admin.png";
    }
    case OP:
    {
        return preferences.m_themePath+"/irc_op.png";
    }
    case VOICE:
    {
        return preferences.m_themePath+"/irc_voice.png";
    }
    case HALFOP:
    {
        return preferences.m_themePath+"/irc_halfop.png";
    }
    default:
    {
        return preferences.m_themePath+"/irc_normal.png";
    }
    }
}

QString dxutils::getUsername()
{
#ifdef Q_OS_WIN
    return QProcessEnvironment::systemEnvironment().value("USERNAME");
#else
    return QProcessEnvironment::systemEnvironment().value("USER");
#endif
}

QString dxutils::getIconForType(TYPE type)
{
    switch(type){
        case SERVER: return getDataPath()+"/icons/world.png";
        case CHANNEL: return getDataPath()+"/icons/channel.png";
        case QUERY: return getDataPath()+"/icons/user.png";
        case OTHER: return getDataPath()+"/icons/logs.png";
        case DCC: return getDataPath()+"/icons/dcc.png";
        case BOATS: return "";
        default: return "";
    }
}

bool dxutils::removeDir(const QString &dirName)
{
    bool result = true;
    QDir dir(dirName);
    if(dir.exists(dirName))
    {
        Q_FOREACH(QFileInfo info, dir.entryInfoList(QDir::NoDotAndDotDot | QDir::System | QDir::Hidden  | QDir::AllDirs | QDir::Files, QDir::DirsFirst))
        {
            if(info.isDir())
            {
                result = removeDir(info.absoluteFilePath());
            }
            else
            {
                result = QFile::remove(info.absoluteFilePath());
            }
            if(!result)
            {
                return result;
            }
        }
        result = dir.rmdir(dirName);
    }
    return result;
}

QString dxutils::getLocalePath()
{
#if defined(Q_OS_WIN32)
    QString path = QApplication::applicationFilePath();
    int ii = path.lastIndexOf('/');
    if(ii > 0)
        path.remove(ii, path.length());
    return path.append("/locale");
#elif defined(Q_OS_DARWIN)
#ifdef ENABLE_OSX_BUNDLE
    QString path = QApplication::applicationFilePath();
    int ii = path.lastIndexOf('/');
    if(ii > 0)
        path.remove(ii, path.length());
    return path.append("/locale");
#else
    return QString(LOCALEDIR);
#endif
#else
    return QString(LOCALEDIR);
#endif
}

QString dxutils::getDataPath()
{
#if defined(Q_OS_WIN32)
    QString path = QApplication::applicationFilePath();
    int ii = path.lastIndexOf('/');
    if(ii > 0)
        path.remove(ii, path.length());
    return path;
#elif defined(Q_OS_DARWIN)
#ifdef ENABLE_OSX_BUNDLE
    QString path = QApplication::applicationFilePath();
    int ii = path.lastIndexOf('/');
    if(ii > 0)
        path.remove(ii, path.length());
    return path.append("/../Resources");
#else
    return QString(DXIRC_DATADIR);
#endif
#else
    return QString(DXIRC_DATADIR);
#endif
}

QKeySequence dxutils::getKeybinding(QKeyEvent *e)
{
    int keyInt = e->key();
    Qt::Key key = static_cast<Qt::Key>(keyInt);
    if(key == Qt::Key_unknown)
    {
        return QKeySequence(Qt::Key_unknown);
    }
    // the user have clicked just and only the special keys Ctrl, Shift, Alt, Meta.
    if(key == Qt::Key_Control ||
        key == Qt::Key_Shift ||
        key == Qt::Key_Alt ||
        key == Qt::Key_Meta)
    {
        return QKeySequence(keyInt);
    }
    // check for a combination of user clicks
    Qt::KeyboardModifiers modifiers = e->modifiers();
    if(modifiers & Qt::ShiftModifier)
        keyInt += Qt::SHIFT;
    if(modifiers & Qt::ControlModifier)
        keyInt += Qt::CTRL;
    if(modifiers & Qt::AltModifier)
        keyInt += Qt::ALT;
    if(modifiers & Qt::MetaModifier)
        keyInt += Qt::META;
    return QKeySequence(keyInt);
}

QString dxutils::wrapText(QString text, int width, char wrap)
{
    QFontMetrics fnt(preferences.m_ircFont);
    int tw = 0;
    int cw;
    int s = -1;
    for(int i=0; i<text.length(); i++)
    {
        cw = fnt.width(text[i]);
        if(text[i]==wrap) s = i;
        if(tw+cw>width)
        {
            if(s!=-1)
            {
                text.replace(s,1,'\n');
                i=s+1;
                s=-1;
            }
            tw=0;
        }
        tw+=cw;
    }
    return text;
}


