/*
 *      tetristabitem.cpp
 *
 *      Copyright 2014 David Vachulka <arch_dvx@users.sourceforge.net>
 *      Copyright (C) 2007-2008 Graeme Gott <graeme@gottcode.org>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */

#include "tetristabitem.h"
#include "config.h"
#include "i18n.h"

static const Cell types[][4] = {
    { Cell(0,0), Cell(0,1), Cell(0,2), Cell(0,3) },
    { Cell(0,0), Cell(0,1), Cell(1,1), Cell(1,2) },
    { Cell(1,0), Cell(0,1), Cell(1,1), Cell(0,2) },
    { Cell(0,0), Cell(0,1), Cell(1,1), Cell(0,2) },
    { Cell(0,0), Cell(1,0), Cell(1,1), Cell(1,2) },
    { Cell(0,0), Cell(1,0), Cell(0,1), Cell(0,2) },
    { Cell(0,0), Cell(1,0), Cell(0,1), Cell(1,1) }
};

static QColor colors[] = {
    QColor(196, 160, 0),
    QColor(206, 92, 0),
    QColor(143, 89, 2),
    QColor(78, 154, 6),
    QColor(32, 74, 135),
    QColor(117, 80, 123),
    QColor(164, 0, 0)
};

static const int timeout = 333;
static const int basicscore = 13;

Piece::Piece(int type, Game *parent)
        : m_type(type), m_parent(parent), m_pivot(4,1), m_valid(false)
{
    Q_ASSERT(type > 0 && type < 8);
    Cell position[4];
    cells(position, type);
    for(int i = 0; i < 4; ++i) position[i].x += 5;
    if(updatePosition(position)) m_valid = true;
}

bool Piece::rotate()
{
    Cell rotated[4];
    for(int i = 0; i < 4; ++i)
    {
        int x = m_cells[i].x - m_pivot.x;
        int y = m_cells[i].y - m_pivot.y;
        rotated[i].x = y + m_pivot.x;
        rotated[i].y = -x + m_pivot.y;
        if(rotated[i].x > 9 || rotated[i].x < 0 || rotated[i].y > 19 || rotated[i].y < 0) return false;
    }
    return updatePosition(rotated);
}

void Piece::drop()
{
    for(int i = 0; i < rows; ++i) moveDown();
}

void Piece::cells(Cell* cells, int type)
{
    Q_ASSERT(cells != 0);
    Q_ASSERT(type > 0 && type < 8);
    const Cell* values = types[type - 1];
    for(int i = 0; i < 4; ++i) cells[i] = values[i];
}

bool Piece::move(int x, int y)
{
    // Move cells
    Cell moved[4];
    for(int i = 0; i < 4; ++i)
    {
        moved[i].x = m_cells[i].x + x;
        moved[i].y = m_cells[i].y + y;
        if(moved[i].x > columns-1 || moved[i].x < 0 || moved[i].y > rows-1 || moved[i].y < 0) return false;
    }

    bool success = updatePosition(moved);
    if (success)
    {
        m_pivot.x += x;
        m_pivot.y += y;
    }
    return success;
}

bool Piece::updatePosition(const Cell* ucells)
{
    // Check for collision of cells
    const Cell* cell = 0;
    bool solid = false;
    for(int i = 0; i < 4; ++i)
    {
        cell = &ucells[i];
        solid = m_parent->cell(cell->x, cell->y);
        if(solid)
        {
            for(int j = 0; j < 4; ++j)
            {
                if(*cell == m_cells[j]) solid = false;
            }
        }
        if(solid) return false;
    }

    // Move cells
    if(m_cells[0].x != -1)
    {
        for(int i = 0; i < 4; ++i)
        {
            m_parent->removeCell(m_cells[i].x, m_cells[i].y);
        }
    }
    for(int i = 0; i < 4; ++i)
    {
        m_parent->addCell(ucells[i].x, ucells[i].y, m_type);
    }
    for(int i = 0; i < 4; ++i)
    {
        m_cells[i] = ucells[i];
    }

    return true;
}

NextPiece::NextPiece(QWidget *parent)
    :   QWidget(parent)
{
    m_anext = qMin(width()/2, height()/4);
    m_piece = 0;
}

void NextPiece::slotPiece(int piece)
{
    m_piece = piece;
    update();
}

void NextPiece::resizeEvent(QResizeEvent *event)
{
    m_anext = qMin(event->size().width()/2, event->size().height()/4);
}

void NextPiece::paintEvent(QPaintEvent *)
{
    if(m_piece<1 || m_piece>7) return;
    QPainter painter(this);
    painter.setRenderHint(QPainter::Antialiasing, true);
    painter.fillRect(QRect(0, 0, width(), height()), QColor("black"));
    for(int i=0; i<4; ++i)
    {
        painter.setBrush(colors[m_piece - 1]);
        painter.drawRect(types[m_piece - 1][i].x*m_anext+1, types[m_piece - 1][i].y*m_anext+1, m_anext-2, m_anext-2);
    }
}

Game::Game(QWidget *parent)
    : QWidget(parent),
      m_removedLines(0),
      m_level(1),
      m_score(0),
      m_nextPiece(0),
      m_paused(false),
      m_done(false),
      m_pauseEnable(false),
      m_piece(0)
{
    setFocusPolicy(Qt::StrongFocus);
    setFocus();
    for(int x=0; x<columns; ++x)
    {
        for(int y=0; y<rows; ++y)
        {
            m_cells[x][y] = 0;
        }
    }
    m_apiece = qMin(width()/columns, height()/rows);
    m_timer = new QTimer(this);
    m_timer->setInterval(timeout);
    m_timer->setSingleShot(true);
    connect(m_timer, SIGNAL(timeout()), this, SLOT(slotTimeout()));
}

bool Game::cell(int x, int y) const
{
    Q_ASSERT(x >= 0 && x < columns);
    Q_ASSERT(y >= 0 && y < rows);
    return m_cells[x][y] != 0;
}

void Game::addCell(int x, int y, int type)
{
    Q_ASSERT(x >= 0 && x < columns);
    Q_ASSERT(y >= 0 && y < rows);
    Q_ASSERT(type > 0 && type < 8);
    Q_ASSERT(m_cells[x][y] == 0);
    m_cells[x][y] = type;
    updateCell(x,y);
}

void Game::removeCell(int x, int y, bool update)
{
    Q_ASSERT(x >= 0 && x < columns);
    Q_ASSERT(y >= 0 && y < rows);
    m_cells[x][y] = 0;
    if(update) updateCell(x,y);
}

void Game::updateCell(int x, int y)
{
    Q_ASSERT(x >= 0 && x < columns);
    Q_ASSERT(y >= 0 && y < rows);
    update();
}

void Game::findFullLines()
{
    // Empty list of full lines
    for(int i = 0; i < 4; ++i)
        m_fullLines[i] = -1;
    int pos = 0;
    // Find full lines
    bool full = false;
    for(int row = 0; row < rows; ++row)
    {
        full = true;
        for(int col = 0; col < columns; ++col)
        {
            if(m_cells[col][row] == 0) full = false;
        }
        if(full)
        {
            m_fullLines[pos] = row;
            ++pos;
        }
    }
}

void Game::newGame()
{
    m_timer->stop();
    delete m_piece;
    m_piece = 0;
    m_paused = false;
    m_done = false;
    m_pauseEnable = true;
    m_removedLines = 0;
    m_level = 1;
    m_score = 0;
    m_timer->setInterval(timeout);
    m_nextPiece = rand()%7+1;
    emit nextPiece(m_nextPiece);
    for(int i=0; i<4; ++i)
        m_fullLines[i] = -1;
    for(int col = 0; col < columns; ++col)
    {
        for(int row = 0; row < rows; ++row)
        {
            m_cells[col][row] = 0;
        }
    }
    createPiece();
    emit started();
    emit updateLabels(m_level, m_score, m_removedLines);
    update();
}

void Game::stopGame()
{
    m_timer->stop();
    delete m_piece;
    m_piece = 0;
    m_paused = false;
    m_done = false;
    m_pauseEnable = false;
    m_removedLines = 0;
    m_level = 1;
    m_score = 0;
    m_nextPiece = rand()%7+1;
    emit nextPiece(m_nextPiece);
    for(int i=0; i<4; ++i)
    m_fullLines[i] = -1;
    for(int col = 0; col < columns; ++col)
    {
        for(int row = 0; row < rows; ++row)
        {
            m_cells[col][row] = 0;
        }
    }
}

void Game::moveLeft()
{
    if(m_done || m_paused) return;
    if(!m_piece) return;
    m_piece->moveLeft();
}

void Game::moveRight()
{
    if(m_done || m_paused) return;
    if(!m_piece) return;
    m_piece->moveRight();
}

void Game::rotate()
{
    if(m_done || m_paused) return;
    if(!m_piece) return;
    m_piece->rotate();
}

void Game::drop()
{
    if(m_done || m_paused) return;
    if(!m_piece) return;
    m_piece->drop();
    landPiece();
}

void Game::createPiece()
{
    Q_ASSERT(m_piece == 0);
    m_piece = new Piece(m_nextPiece, this);
    if(m_piece->isValid())
    {
        m_nextPiece = rand()%7+1;
        emit nextPiece(m_nextPiece);
        m_timer->start();
    }
    else
    {
        delete m_piece;
        m_piece = 0;
        m_done = true;
        gameOver();
        return;
    }
    update();
}

void Game::removeLines()
{
    int deltascore = basicscore*m_level;
    // Loop through full lines
    for(int i = 0; i < 4; ++i)
    {
        int row = m_fullLines[i];
        if(row == -1)
            break;
        // Remove line
        for(int col = 0; col < 10; ++col)
        {
            removeCell(col, row, false); //not need update canvas, CreatePiece do it
        }
        ++m_removedLines;
        deltascore *= 3;
        // Shift board down
        for(; row > 0; --row)
        {
            for(int col = 0; col < columns; ++col)
            {
                m_cells[col][row] = m_cells[col][row - 1];
            }
        }
    }
    // Remove top line
    if(m_fullLines[0] != -1)
    {
        for(int col = 0; col < 10; ++col)
        {
            removeCell(col, 0, false); //not need update canvas, CreatePiece do it
        }
    }
    m_level = (m_removedLines / 10) + 1;
    m_score += deltascore;
    m_timer->setInterval(timeout-(m_level-1)*13);
    // Empty list of full lines
    for(int i = 0; i < 4; ++i)
        m_fullLines[i] = -1;
    emit updateLabels(m_level, m_score, m_removedLines);
    // Add new piece
    createPiece();
}

void Game::landPiece()
{
    m_timer->stop();
    delete m_piece;
    m_piece = 0;
    findFullLines();
    if(m_fullLines[0] != -1) removeLines();
    else createPiece();
}

void Game::pauseResumeGame()
{
    if(m_done) return;
    if(!m_pauseEnable) return;
    if(m_paused)
    {
        m_timer->start();
    }
    else
    {
        m_timer->stop();
    }
    m_paused = !m_paused;
    emit pausedResumed();
    update();
}

void Game::gameOver()
{
    m_timer->stop();
    m_done = true;
    m_pauseEnable = false;
    emit finished();
    update();
}

void Game::paintEvent(QPaintEvent*)
{
    QPainter painter(this);
    painter.setRenderHint(QPainter::Antialiasing, true);
    painter.fillRect(QRect(0, 0, width(), height()), QColor("black"));
    painter.setPen(QPen(QColor("white"),2));
    QFont font;
    QFontMetrics fontMetrics(font);
    painter.drawLine(0,rows*m_apiece,columns*m_apiece,rows*m_apiece);
    painter.drawLine(columns*m_apiece,rows*m_apiece,columns*m_apiece,0);
    painter.setPen(QColor("black"));
    for(int x=0; x<columns; ++x)
    {
        for(int y=0; y<rows; ++y)
        {
            if(m_cells[x][y] != 0)
            {
                painter.setBrush(colors[m_cells[x][y]-1]);
                painter.drawRect(x*m_apiece, y*m_apiece, m_apiece-2, m_apiece-2);
            }
        }
    }
    painter.setPen(QColor("white"));
    if(!m_done && !m_piece)
    {
        painter.setFont(font);
        painter.drawText(2, fontMetrics.height()+2, _("Keys for playing:"));
        painter.drawText(2, 2*(fontMetrics.height()+2), _("n .. new game"));
        painter.drawText(2, 3*(fontMetrics.height()+2), _("p .. pause game"));
        painter.drawText(2, 4*(fontMetrics.height()+2), _("i .. rotate piece"));
        painter.drawText(2, 5*(fontMetrics.height()+2), _("l .. move piece right"));
        painter.drawText(2, 6*(fontMetrics.height()+2), _("k .. drop piece"));
        painter.drawText(2, 7*(fontMetrics.height()+2), _("j .. move piece left"));
    }
    font.setPointSize(25);
    font.setWeight(QFont::Bold);
    painter.setFont(font);
    fontMetrics = QFontMetrics(font);
    if(m_done)
    {
        int width = fontMetrics.width(_("GAME OVER"));
        int x = 2;
        if(width < m_apiece*columns-4) x = (m_apiece*columns-width)/2;
        painter.drawText(x, height()/2, _("GAME OVER"));
    }
    if(m_paused)
    {
        int width = fontMetrics.width(_("PAUSED"));
        int x = 2;
        if(width < m_apiece*columns-4) x = (m_apiece*columns-width)/2;
        painter.drawText(x, height()/2, _("PAUSED"));
    }
}

void Game::slotTimeout()
{
    Q_ASSERT(m_piece != 0);
    if(m_piece->moveDown())
    {
        update();
        m_timer->start();
    }
    else landPiece();
}

void Game::slotNewGame()
{
    newGame();
}

void Game::slotPauseGame()
{
    pauseResumeGame();
}

void Game::keyPressEvent(QKeyEvent* event)
{
    switch(event->key()) {
    case Qt::Key_N:
        newGame();
        break;
    case Qt::Key_P:
        pauseResumeGame();
        break;
    case Qt::Key_J:
        moveLeft();
        break;
    case Qt::Key_L:
        moveRight();
        break;
    case Qt::Key_I:
        rotate();
        break;
    case Qt::Key_K:
        drop();
        break;
    default:
        break;
    }
    update();
}

void Game::mousePressEvent(QMouseEvent* event)
{
    if(m_paused)
    {
        pauseResumeGame();
    }
    else if(m_done)
    {
        newGame();
    }
    QWidget::mousePressEvent(event);
}

void Game::focusOutEvent(QFocusEvent*)
{
    if(m_piece && !m_done && !m_paused)
        pauseResumeGame();
}

void Game::resizeEvent(QResizeEvent* event)
{
    m_apiece = qMin(event->size().width()/columns, event->size().height()/rows);
}

TetrisTabItem::TetrisTabItem(TabWidget *tabbook, int id, QWidget *parent)
:   QWidget(parent),
    m_parent(tabbook),
    m_id(id)
{
    m_splitter = new QSplitter(this);
    m_splitter->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);

    m_gamecanvas = new Game(this);
    m_gamecanvas->setFocus();
    m_splitter->addWidget(m_gamecanvas);

    QWidget *other = new QWidget(this);
    QVBoxLayout *frame = new QVBoxLayout;
    frame->addWidget(new QLabel(_("Next piece:")));
    m_nextcanvas = new NextPiece(this);
    m_nextcanvas->setFixedWidth(40);
    m_nextcanvas->setFixedHeight(80);
    frame->addWidget(m_nextcanvas);
    m_levelLabel = new QLabel(_("Level: %1").arg(1));
    frame->addWidget(m_levelLabel);
    m_scoreLabel = new QLabel(_("Score: %1").arg(0));
    frame->addWidget(m_scoreLabel);
    m_linesLabel = new QLabel(_("Lines: %1").arg(0));
    frame->addWidget(m_linesLabel);
    m_newButton = new QPushButton(_("&New game"));
    frame->addWidget(m_newButton);
    m_pauseButton = new QPushButton(_("&Pause game"));
    m_pauseButton->setEnabled(false);
    frame->addWidget(m_pauseButton);
    frame->addStretch(1);
    other->setLayout(frame);
    m_splitter->addWidget(other);

    QVBoxLayout *mainLayout = new QVBoxLayout(this);
    mainLayout->addWidget(m_splitter);
    setLayout(mainLayout);
    QList<int> list;
    list << width()-120 << 120;
    m_splitter->setSizes(list);

    connect(m_gamecanvas, SIGNAL(nextPiece(int)), m_nextcanvas, SLOT(slotPiece(int)));
    connect(m_gamecanvas, SIGNAL(updateLabels(int,int,int)), this, SLOT(slotUpdateLabels(int,int,int)));
    connect(m_gamecanvas, SIGNAL(started()), this, SLOT(slotNewGame()));
    connect(m_gamecanvas, SIGNAL(pausedResumed()), this, SLOT(slotPauseGame()));
    connect(m_gamecanvas, SIGNAL(finished()), this, SLOT(slotGameOver()));
    connect(m_newButton, SIGNAL(clicked()), m_gamecanvas, SLOT(slotNewGame()));
    connect(m_pauseButton, SIGNAL(clicked()), m_gamecanvas, SLOT(slotPauseGame()));
}

TetrisTabItem::~TetrisTabItem()
{
}

void TetrisTabItem::slotUpdateLabels(int level, int score, int removedLines)
{
    m_levelLabel->setText(_("Level: %1").arg(level));
    m_scoreLabel->setText(_("Score: %1").arg(score));
    m_linesLabel->setText(_("Lines: %1").arg(removedLines));
}

void TetrisTabItem::slotNewGame()
{
    m_newButton->setEnabled(false);
    m_pauseButton->setEnabled(true);
    m_pauseButton->setText(_("&Pause game"));
}

void TetrisTabItem::slotPauseGame()
{
    if(!m_gamecanvas->isPauseEnable()) return;
    if(!m_gamecanvas->isPaused())
    {
        m_pauseButton->setText(_("&Pause game"));
    }
    else
    {
        m_newButton->setEnabled(true);
        m_pauseButton->setText(_("&Resume game"));
    }
}

void TetrisTabItem::slotGameOver()
{
    m_newButton->setEnabled(true);
    m_pauseButton->setEnabled(false);
}

void TetrisTabItem::setGameFocus()
{
    m_gamecanvas->setFocus();
}

